from django.contrib import admin
from django.utils.safestring import mark_safe

from .utils import new_session
from . import models
from .chessgames import get_best_match, update_video
from pprint import pformat


class VideoAdmin(admin.ModelAdmin):
    date_hierarchy = 'time_published'
    list_display = [
        'id',
        'title',
        'time_published',
        'white_name',
        'black_name',
        'chessgames_id',
        'stream',
        'engine',
        'ignore',

    ]
    list_filter = [
        'ignore',
        'stream',
        'engine',
    ]
    list_editable = [

        'ignore',
        'stream',
        'engine',
    ]
    search_fields = [
        'id',
        'title',
        'description',
        'youtube_id',
    ]
    ordering = [
        '-time_published'
    ]
    readonly_fields = [
        'time_updated',
        'extracted_moves',
        'extracted_players',
        'chessgames_similar_game_search_url',
        'data_pre',
        'search_vector',
    ]

    actions = [
        'inline_sync_with_api',
        'schedule_sync_with_api',
        'get_best_match',
        'update_from_chessgames',
        'match_on_similarity',
    ]

    # class Media:
    #     js = (
    #         'agadmator/chessgames.js',
    #     )

    def get_best_match(self, r, qs):
        session = new_session()
        success = 0
        for obj in qs:
            result_id = get_best_match(obj, session=session)
            if result_id:
                success += 1
            update_video(obj, session=session)
        self.message_user(r, f'Auto match: Total {len(qs)}, Success {success}')

    def match_on_similarity(self, r, qs):
        success = 0
        for obj in qs:
            result = obj.match_on_possible_game_similarity()
            if result:
                success += 1
        self.message_user(r, f'Similarity matching: Total {len(qs)}, Success {success}')

    def update_from_chessgames(self, r, qs):
        session = new_session()
        for obj in qs:
            update_video(obj, session=session)
        self.message_user(r, f'Update: Total {len(qs)}')

    def data_pre(self, obj):
        return mark_safe(f'<pre>{pformat(obj.data, indent=2)}</pre>')


admin.site.register(models.Video, VideoAdmin)



from urllib.parse import urlparse, parse_qs


def chessgames_player(html_element):
    root = html_element
    name = html_element.cssselect('p b')[0].text
    tables = html_element.cssselect('table')
    try:
        description = tables[10].cssselect('p')[0].text
    except IndexError:
        description = ''

    return {
        'name': name,
        'description': description
    }


def player_directory(html_element):
    anchors = html_element.cssselect('a')
    players = []
    for a in anchors:
        href = a.get('href')
        if not href.startswith('/perl/chessplayer'):
            continue
        try:
            pid = parse_qs(urlparse(href).query).get('pid')[0]
        except (IndexError, TypeError):
            continue

        players.append({
            'url': href,
            'name': a.text_content(),
            'id': pid
        })

    return {
        'players': players
    }



import chess
from chess.pgn import StringExporter


class FenExporter(StringExporter):

    def __init__(self):
        super().__init__(comments=False, headers=False, variations=False, columns=None)
        self.positions = []

    def visit_header(self, tagname, tagvalue):
        pass

    def visit_move(self, board, move):
        self.positions.append(board.fen())

    def result(self):
        return self.positions


class MarkupExporter(StringExporter):

    def __init__(self):
        super().__init__(columns=None, headers=False, comments=False, variations=False)

    def visit_move(self, board, move):
        if self.variations or not self.variation_depth:
            # Write the move number.
            if board.turn == chess.WHITE:

                self.write_token(str(board.fullmove_number) + ". ")
            elif self.force_movenumber:
                self.write_token(str(board.fullmove_number) + "... ")

            offset = -1 if board.turn == chess.WHITE else 0
            n = 2 * board.fullmove_number + offset
            # Write the SAN.
            self.write_token(f'<span class="cp" id="p-{n}" data-i="{n}">{board.san(move)}</span> ')
            self.force_movenumber = False

    def visit_result(self, result):
        pass


from django.urls import path
from agadmator import views

urlpatterns = [
    path('videos/', views.VideoListView.as_view(), name='video_list'),
    path('videos/<str:slug>/', views.VideoDetailView.as_view(), name='video_detail'),
    path('player/<str:id_or_slug>/', views.PlayerDetailView.as_view(), name='player_detail'),
    path('tournament/<int:chessgames_id>/', views.TournamentDetailView.as_view(), name='tournament_detail'),
    path('year/<int:year>/', views.YearDetailView.as_view(), name='year_detail'),
    path('published/<year_month>/', views.PublishedMonthDetailView.as_view(), name='published_month_detail'),
    path('about/', views.AboutView.as_view(), name='about'),
    path('', views.IndexView.as_view())
]

app_name = 'agadmator'


# Generated by Django 2.1.3 on 2018-11-05 16:35

from django.db import migrations
from agadmator.models import Video


def populate_values(apps, schema_editor):

    for vid in Video.objects.all():
        vid.white_name = vid.get_white_name
        vid.white_slug = vid.get_white_slug
        vid.black_name = vid.get_black_name
        vid.black_slug = vid.get_black_slug
        vid.tournament_id = vid.get_tournament_id
        vid.tournament_name = vid.get_tournament_name
        vid.year = vid.get_year
        vid.save()


class Migration(migrations.Migration):
    dependencies = [
        ('agadmator', '0004_auto_20181105_1628'),
    ]

    operations = [
        migrations.RunPython(
            populate_values, reverse_code=migrations.RunPython.noop
        )
    ]



# Generated by Django 2.1.3 on 2018-11-05 07:07

import django.contrib.postgres.fields.jsonb
import django.contrib.postgres.indexes
import django.contrib.postgres.search
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Video',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('youtube_id', models.CharField(blank=True, db_index=True, max_length=100)),
                ('title', models.CharField(blank=True, max_length=500)),
                ('time_updated', models.DateTimeField(auto_now=True)),
                ('time_published', models.DateTimeField(blank=True, db_index=True, null=True)),
                ('date_played', models.DateField(blank=True, db_index=True, null=True)),
                ('description', models.TextField(blank=True)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('chessgames_id', models.BigIntegerField(blank=True, db_index=True, null=True)),
                ('search_vector', django.contrib.postgres.search.SearchVectorField(null=True)),
            ],
            options={
                'verbose_name': 'Video',
                'verbose_name_plural': 'Videos',
            },
        ),
        migrations.AddIndex(
            model_name='video',
            index=django.contrib.postgres.indexes.GinIndex(fields=['search_vector'], name='agadmator_v_search__8d5e27_gin'),
        ),
    ]



import logging
import re
from urllib.parse import urlencode

from django.contrib.postgres.fields import JSONField
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVectorField
from django.db import models
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.text import slugify
from .utils import extract_pgn, extract_players
from chess.pgn import read_game, StringExporter
from io import StringIO
from .exporters import MarkupExporter, FenExporter

logger = logging.getLogger(__name__)


class Video(models.Model):
    youtube_id = models.CharField(blank=True, max_length=100, db_index=True)
    title = models.CharField(max_length=500, blank=True)
    time_updated = models.DateTimeField(auto_now=True)
    time_published = models.DateTimeField(blank=True, null=True, db_index=True)
    date_played = models.DateField(blank=True, null=True, db_index=True)
    description = models.TextField(blank=True)
    data = JSONField(default=dict, blank=True)
    chessgames_id = models.BigIntegerField(blank=True, null=True, db_index=True)
    search_vector = SearchVectorField(null=True)
    stream = models.BooleanField(default=False, db_index=True)  # if video is YT stream
    engine = models.BooleanField(default=False, db_index=True)   # if one of the players is engine
    ignore = models.BooleanField(default=False, db_index=True)  # if video is not fit for normal parsing
    error = models.BooleanField(default=False, db_index=True)  # if there was an error during matching
    error_message = models.TextField(blank=True)
    white_name = models.CharField(max_length=500, blank=True)
    white_slug = models.SlugField(max_length=500, blank=True, db_index=True)
    black_name = models.CharField(max_length=500, blank=True)
    black_slug = models.CharField(max_length=500, blank=True, db_index=True)
    tournament_id = models.BigIntegerField(blank=True, null=True, db_index=True)
    tournament_name = models.CharField(max_length=500, blank=True)
    year = models.PositiveIntegerField(blank=True, null=True)  # year in which game was played
    """
    data keys:
       
        white: {
            name: str
            slug: str
            chessgames_id: str
        }
        black: {
            name: str
            slug: str
            chessgames_id: str
        }
        chessgames_id_list: [str]  // used for auto game matching
        chessgames_success: bool    // if auto matching succeeded
        chessgames_pgn: str
        tournament: {
            name: str
            chessgames_id: str
        }                   // name of tournament this game is part of
        duration: int       // video duration in seconds (from youtube_dl)
        tags: [str]         // tags from youtube
        
    """

    def __str__(self):
        return self.title or self.youtube_id

    def get_absolute_url(self):
        return reverse('agadmator:video_detail', args=[self.youtube_id])

    def modify(self, save=True):
        # agadmator line separator
        self.description = re.sub(r'[-]{5,}', '---', self.description)
        if save:
            self.save()

    def store_names(self):

        white, black = self.extracted_players

        if not self.data.get('white'):
            self.data['white'] = {
                'name': white,
                'slug': slugify(white),
            }

        if not self.data.get('black'):
            self.data['black'] = {
                'name': black,
                'slug': slugify(black)
            }

        self.save(update_fields=['data'])

    @cached_property
    def get_white_name(self):
        white = self.data.get('white')
        return white.get('name') if white else self.extracted_players[0]

    @cached_property
    def get_white_slug(self):
        white = self.data.get('white')
        return white.get('slug') if white else slugify(self.get_white_name)

    @cached_property
    def get_black_name(self):
        black = self.data.get('black')
        return black.get('name') if black else self.extracted_players[1]

    @cached_property
    def get_black_slug(self):
        black = self.data.get('black')
        return black.get('slug') if black else slugify(self.get_black_name)

    @cached_property
    def get_tournament_id(self):
        t = self.data.get('tournament')
        if t:
            return t.get('chessgames_id')
        return None

    @cached_property
    def get_tournament_name(self):
        t = self.data.get('tournament')
        if t:
            return t.get('name')
        return ''

    @cached_property
    def get_year(self):
        return self.data.get('year')

    def get_test(self):
        return "XXX TEST XXXX"

    @property
    def thumbnail_default(self):
        return f'https://i.ytimg.com/vi/{self.youtube_id}/default.jpg'

    @property
    def thumbnail_medium(self):
        return f'https://i.ytimg.com/vi/{self.youtube_id}/mqdefault.jpg'

    @property
    def thumbnail_high(self):
        return f'https://i.ytimg.com/vi/{self.youtube_id}/hqdefault.jpg'

    @cached_property
    def extracted_pgn(self):
        return extract_pgn(self.description)

    @cached_property
    def extracted_players(self):
        return extract_players(self.description)

    @cached_property
    def extracted_moves(self):
        numbers = re.findall(r'(\d+)\.', self.extracted_pgn)
        if numbers:
            return numbers[-1]
        return None


    @cached_property
    def pgn(self):
        chessgames_pgn = self.data.get('chessgames_pgn')
        return chessgames_pgn if chessgames_pgn else self.extracted_pgn

    @cached_property
    def normalized_pgn(self):
        if not self.game:
            return ''
        exporter = StringExporter(
            columns=None, headers=False, comments=False, variations=False
        )
        return self.game.accept(exporter).rstrip('* ')

    @cached_property
    def fen_list(self):
        if not self.game:
            return ''
        exporter = FenExporter()
        positions = self.game.accept(exporter)
        positions.append(self.game.end().board().fen())
        return positions

    @cached_property
    def pgn_markup(self):
        if not self.game:
            return ''
        exporter = MarkupExporter()
        return self.game.accept(exporter)

    @cached_property
    def game(self):
        if not self.pgn:
            return None
        return read_game(StringIO(self.pgn))

    @property
    def chessgames_similar_game_search_url(self):

        player, player2 = self.extracted_players

        params = {
            'yearcomp': 'exactly',
            'year': '',
            'playercomp': 'either',
            'player': player,
            'player2': player2,
            'movescomp': 'exactly',
            'moves': self.extracted_moves,
        }

        return 'http://www.chessgames.com/perl/chess.pl?' + urlencode(params)

    class Meta:
        verbose_name = 'Video'
        verbose_name_plural = 'Videos'
        indexes = [
            GinIndex(fields=['search_vector'])
        ]

    def tags(self):
        tags = []

        if self.get_white_slug:
            tags.append({
                'link': reverse('agadmator:player_detail', args=[self.get_white_slug]),
                'text': self.get_white_name
            })

        if self.get_black_slug:
            tags.append({
                'link': reverse('agadmator:player_detail', args=[self.get_black_slug]),
                'text': self.get_black_name
            })

        tournament = self.data.get('tournament')

        if tournament:
            tournament_name = tournament.get('name')
            tournament_id = tournament.get('chessgames_id')
            tags.append({
                'link': reverse('agadmator:tournament_detail', args=[tournament_id]),
                'text': tournament_name
            })

        return tags



from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def set_qsarg(context, *args):

    query_dict = context['request'].GET.copy()
    for key, val in zip(args[::2], args[1::2]):
        query_dict[str(key)] = str(val)
    return '?' + query_dict.urlencode()



from django.test import TestCase, SimpleTestCase
import pathlib
# from .models import Player, Tournament, Game
from .utils import clean_pgn, extract_pgn, date_from_pgn_date_header, parse_pgn
from . import chessgames


class TestDataExtraction(TestCase):

    def test_game(self):
        g = Game(id=123)
        html = pathlib.Path(__file__, '../tests/game.html').resolve().read_text()
        g.update(markup=html)
        self.assertEqual(g.white_id, 12295)
        self.assertEqual(g.black_id, 50065)
        self.assertEqual(g.tournament_id, 93549)
        self.assertEqual(g.eco, 'A14')
        self.assertEqual(g.pgn, """[Event "Grand Chess Tour Paris (Rapid)"]
[Site "Paris FRA"]
[Date "2018.06.20"]
[EventDate "2018.06.20"]
[Round "3"]
[Result "1-0"]
[White "Vladimir Kramnik"]
[Black "Shakhriyar Mamedyarov"]
[ECO "A14"]
[WhiteElo "2792"]
[BlackElo "2808"]
[PlyCount "98"]

1. Nf3 Nf6 2. g3 d5 3. Bg2 e6 4. O-O Be7 5. c4 O-O 6. b3 c5 7. cxd5 Nxd5 8. Bb2 Nc6 9. d4 cxd4 10. Nxd4 Nxd4 11. Qxd4 Bf6 12. Qd2 Nf4 13. gxf4 Qxd2 14. Nxd2 Bxb2 15. Rab1 Bf6 16. Rbc1 Rd8 17. Ne4 Be7 18. Rc7 Kf8 19. Rfc1 Rb8 20. e3 h6 21. Bf3 a5 22. Kg2 g5 23. fxg5 hxg5 24. h3 f5 25. Nc5 b6 26. Na4 Bd6 27. Rh7 Kg8 28. Ra7 Be5 29. Rc6 Rd6 30. Rxd6 Bxd6 31. Nc3 b5 32. Rxa5 b4 33. Nb5 Bc5 34. Nd4 Bxd4 35. exd4 Kf7 36. Ra7+ Kf6 37. Rc7 Ba6 38. d5 Bb7 39. d6 Bxf3+ 40. Kxf3 Rd8 41. Rb7 Rxd6 42. Rxb4 e5 43. a4 e4+ 44. Ke2 Ke5 45. a5 f4 46. Ra4 f3+ 47. Ke1 Rc6 48. a6 Rc1+ 49. Kd2 Rf1 1-0""")

    def test_tournament(self):
        html = pathlib.Path(__file__, '../tests/tournament.html').resolve().read_text(errors='ignore')
        t = Tournament(id=123)
        t.update(markup=html)
        self.assertEqual(t.name, 'Grand Chess Tour Paris (Rapid) Tournament')

    def test_player(self):
        html = pathlib.Path(__file__, '../tests/player.html').resolve().read_text(errors='ignore')
        p = Player(id=123)
        p.update(markup=html)
        self.assertEqual(p.name, 'Vladimir Kramnik')


class TestUtils(SimpleTestCase):

    def test_clean_pgn(self):
        s = """
[Event "Biel"]
[Site "0:52:33-0:22:33"]
[Date "2018.07.22"]
[EventDate "2018.07.21"]
[Round "1"]
[Result "1-0"]
[White "Magnus Carlsen"]
[Black "David Navara"]
[ECO "D30"]
[WhiteElo "?"]
[BlackElo "?"]
[PlyCount "127"]

1.d4 Nf6 2.Nf3 d5 3.c4 e6 4.Bg5 Bb4+ 5.Nc3 h6 6.Bxf6 Qxf6 7.e3 O-O 8.Rc1 dxc4 9.Bxc4 c5 10.dxc5 Nd7 11.O-O Nxc5 12.Nb5 a6 13.Nbd4 b5 14.Be2 e5 15.Nc2 Rd8 16.Nxb4 Rxd1 17.Rfxd1 a5 18.Nd5 Qd6 19.Nxe5 Bb7 20.Bf3 Rc8 21.Ng4 Qf8 22.h4 Nd7 23.Rxc8 Bxc8 24.a3 h5 25.Nh2 g6 26.Be2 Ne5 27.Bxb5 Bb7 28.Nc3 Qe7 29.Rd4 Qe6 30.Nf1 Qb3 31.Rd2 Nc4 32.Rd7 Nxb2 33.Rxb7 Qxc3 34.Be8 Kf8 35.Bxf7 Qc6 36.Rxb2 Kxf7 37.Rd2 Qa4 38.Rd3 Qxh4 39.Rd7+ Kg8 40.Rd4 Qe7 41.a4 Qa3 42.g3 Qa1 43.Kg2 g5 44.Nd2 g4 45.Ne4 Qc1 46.Nf6+ Kf7 47.Nxh5 Qc6+ 48.Kg1 Qc1+ 49.Kh2 Kg6 50.Nf4+ Kf6 51.Ng2 Kg5 52.Rf4 Qd1 53.Nh4 Qc2 54.Nf5 Qd3 55.e4 Qd7 56.e5 Qh7+ 57.Kg1 Qg6 58.Nd6 Qe6 59.Rf5+ Qxf5 60.Nxf5 Kxf5 61.f4 gxf3 62.Kf2 Kxe5 63.Kxf3 Kf5 64.Ke3 1-0
        """

        expected = """[Event "Biel"]
[Site "0:52:33-0:22:33"]
[Date "2018.07.22"]
[EventDate "2018.07.21"]
[Round "1"]
[Result "1-0"]
[White "Magnus Carlsen"]
[Black "David Navara"]
[ECO "D30"]
[PlyCount "127"]

1.d4 Nf6 2.Nf3 d5 3.c4 e6 4.Bg5 Bb4+ 5.Nc3 h6 6.Bxf6 Qxf6 7.e3 O-O 8.Rc1 dxc4 9.Bxc4 c5 10.dxc5 Nd7 11.O-O Nxc5 12.Nb5 a6 13.Nbd4 b5 14.Be2 e5 15.Nc2 Rd8 16.Nxb4 Rxd1 17.Rfxd1 a5 18.Nd5 Qd6 19.Nxe5 Bb7 20.Bf3 Rc8 21.Ng4 Qf8 22.h4 Nd7 23.Rxc8 Bxc8 24.a3 h5 25.Nh2 g6 26.Be2 Ne5 27.Bxb5 Bb7 28.Nc3 Qe7 29.Rd4 Qe6 30.Nf1 Qb3 31.Rd2 Nc4 32.Rd7 Nxb2 33.Rxb7 Qxc3 34.Be8 Kf8 35.Bxf7 Qc6 36.Rxb2 Kxf7 37.Rd2 Qa4 38.Rd3 Qxh4 39.Rd7+ Kg8 40.Rd4 Qe7 41.a4 Qa3 42.g3 Qa1 43.Kg2 g5 44.Nd2 g4 45.Ne4 Qc1 46.Nf6+ Kf7 47.Nxh5 Qc6+ 48.Kg1 Qc1+ 49.Kh2 Kg6 50.Nf4+ Kf6 51.Ng2 Kg5 52.Rf4 Qd1 53.Nh4 Qc2 54.Nf5 Qd3 55.e4 Qd7 56.e5 Qh7+ 57.Kg1 Qg6 58.Nd6 Qe6 59.Rf5+ Qxf5 60.Nxf5 Kxf5 61.f4 gxf3 62.Kf2 Kxe5 63.Kxf3 Kf5 64.Ke3 1-0"""

        self.assertEqual(clean_pgn(s), expected)

    def test_extract_pgn(self):
        text = """Robert James Fischer vs Borislav Ivkov
Palma de Mallorca Interzonal (1970), Palma de Mallorca ESP, rd 13, Nov-27
Spanish Game: Closed Variations. Smyslov Defense (C93)

1. e4 e5 2. Nf3 Nc6 3. Bb5 a6 4. Ba4 Nf6 5. O-O Be7 6. Re1 b5 7. Bb3 d6 8. c3 O-O 9. h3 h6 10. d4 Re8 11. Be3 Bf8 12. Nbd2 Bb7 13. Qb1 Qb8 14. a3 Nd8 15. Bc2 c6 16. b4 Qc7 17. Bd3 Ne6 18. Qc2 Rac8 19. a4 Nd7 20. Qb3 Nf4 21. Bf4 ef4 22. c4 bc4 23. Bc4 d5 24. Bd3 Rb8 25. Qc3 Qb6 26. Reb1 Ba8 27. a5 Qa7 28. e5 Rb7 29. Qc6 Bb4 30. Qa6 Qb8 31. Bc2 Re6 32. Qd3 Nf8 33. Qf5 Ra6 34. Nb3 g6 35. Qf4 Bc3 36. Nc5 Ra5 37. Nb7 Ra1 38. Ra1 Ne6 39. Qf6 Ba1 40. Nd6 Qc7 41. Bg6 Qc1 42. Ne1 Qe1 43. Kh2 Ng5 44. Bf7
---
If you realllly enjoy my content, you're welcome to support me and my channel with a small donation via PayPal, Bitcoin, Litecoin or Nano."""
        expected = """1. e4 e5 2. Nf3 Nc6 3. Bb5 a6 4. Ba4 Nf6 5. O-O Be7 6. Re1 b5 7. Bb3 d6 8. c3 O-O 9. h3 h6 10. d4 Re8 11. Be3 Bf8 12. Nbd2 Bb7 13. Qb1 Qb8 14. a3 Nd8 15. Bc2 c6 16. b4 Qc7 17. Bd3 Ne6 18. Qc2 Rac8 19. a4 Nd7 20. Qb3 Nf4 21. Bf4 ef4 22. c4 bc4 23. Bc4 d5 24. Bd3 Rb8 25. Qc3 Qb6 26. Reb1 Ba8 27. a5 Qa7 28. e5 Rb7 29. Qc6 Bb4 30. Qa6 Qb8 31. Bc2 Re6 32. Qd3 Nf8 33. Qf5 Ra6 34. Nb3 g6 35. Qf4 Bc3 36. Nc5 Ra5 37. Nb7 Ra1 38. Ra1 Ne6 39. Qf6 Ba1 40. Nd6 Qc7 41. Bg6 Qc1 42. Ne1 Qe1 43. Kh2 Ng5 44. Bf7"""
        self.assertEqual(extract_pgn(text), expected)

    def test_date_from_pgn(self):
        pgn = """
[Event "Biel"]
[Site "0:52:33-0:22:33"]
[Date "2018.07.22"]
[EventDate "2018.07.21"]
[Round "1"]
[Result "1-0"]
[White "Magnus Carlsen"]
[Black "David Navara"]
[ECO "D30"]
[WhiteElo "?"]
[BlackElo "?"]
[PlyCount "127"]

1.d4 Nf6 2.Nf3 d5 3.c4 e6 4.Bg5 Bb4+ 5.Nc3 h6 6.Bxf6 Qxf6 7.e3 O-O 8.Rc1 dxc4 9.Bxc4 c5 10.dxc5 Nd7 11.O-O Nxc5 12.Nb5 a6 13.Nbd4 b5 14.Be2 e5 15.Nc2 Rd8 16.Nxb4 Rxd1 17.Rfxd1 a5 18.Nd5 Qd6 19.Nxe5 Bb7 20.Bf3 Rc8 21.Ng4 Qf8 22.h4 Nd7 23.Rxc8 Bxc8 24.a3 h5 25.Nh2 g6 26.Be2 Ne5 27.Bxb5 Bb7 28.Nc3 Qe7 29.Rd4 Qe6 30.Nf1 Qb3 31.Rd2 Nc4 32.Rd7 Nxb2 33.Rxb7 Qxc3 34.Be8 Kf8 35.Bxf7 Qc6 36.Rxb2 Kxf7 37.Rd2 Qa4 38.Rd3 Qxh4 39.Rd7+ Kg8 40.Rd4 Qe7 41.a4 Qa3 42.g3 Qa1 43.Kg2 g5 44.Nd2 g4 45.Ne4 Qc1 46.Nf6+ Kf7 47.Nxh5 Qc6+ 48.Kg1 Qc1+ 49.Kh2 Kg6 50.Nf4+ Kf6 51.Ng2 Kg5 52.Rf4 Qd1 53.Nh4 Qc2 54.Nf5 Qd3 55.e4 Qd7 56.e5 Qh7+ 57.Kg1 Qg6 58.Nd6 Qe6 59.Rf5+ Qxf5 60.Nxf5 Kxf5 61.f4 gxf3 62.Kf2 Kxe5 63.Kxf3 Kf5 64.Ke3 1-0
        """

        expected = '2018-07-22'

        self.assertEqual(date_from_pgn(pgn).strftime('%Y-%m-%d'), expected)


class TestChessgamesDataExtraction(SimpleTestCase):

    def test_game(self):
        markup = pathlib.Path(__file__, '../tests/chessgames_game.html').resolve().read_text(errors='ignore')

        data = chessgames.get_game(game_id=123, markup=markup)

        white = {
            'chessgames_id': 13741,
            'name': 'Lev Polugaevsky'
        }

        tournament = {
            'chessgames_id': 79537,
            'name': 'USSR Championship 1961a (1961)'
        }

        self.assertEqual(data['year'], 1961)
        self.assertEqual(data['headers']['ECO'], 'C98')
        self.assertDictEqual(data['white'], white)
        self.assertDictEqual(data['tournament'], tournament)

    def test_pgn_headers(self):
        pgn = """
        [Event "Biel"]
        [Site "0:52:33-0:22:33"]
        [Date "2018.07.22"]
        [EventDate "2018.07.21"]
        [Round "1"]
        [Result "1-0"]
        [White "Magnus Carlsen"]
        [Black "David Navara"]
        [ECO "D30"]
        [WhiteElo "?"]
        [BlackElo "?"]
        [PlyCount "127"]
        
        xxx
        """

        headers = parse_pgn(pgn).headers
        expected = {
            'Event': 'Biel',
            'Site': '0:52:33-0:22:33',
            'Date': '2018.07.22',
            'EventDate': '2018.07.21',
            'Round': '1',
            'Result': '1-0',
            'White': 'Magnus Carlsen',
            'Black': 'David Navara',
            'ECO': 'D30',
            'WhiteElo': '?',
            'BlackElo': '?',
            'PlyCount': '127',
        }
        self.assertDictEqual(headers, expected)



# Generated by Django 2.1.3 on 2018-11-05 16:28

from django.db import migrations, models




class Migration(migrations.Migration):
    dependencies = [
        ('agadmator', '0003_auto_20181105_1613'),
    ]

    operations = [
        migrations.AddField(
            model_name='video',
            name='black_name',
            field=models.CharField(blank=True, max_length=500),
        ),
        migrations.AddField(
            model_name='video',
            name='black_slug',
            field=models.CharField(blank=True, db_index=True, max_length=500),
        ),
        migrations.AddField(
            model_name='video',
            name='error',
            field=models.BooleanField(db_index=True, default=False),
        ),
        migrations.AddField(
            model_name='video',
            name='tournament_id',
            field=models.BigIntegerField(blank=True, db_index=True, null=True),
        ),
        migrations.AddField(
            model_name='video',
            name='tournament_name',
            field=models.CharField(blank=True, max_length=500),
        ),
        migrations.AddField(
            model_name='video',
            name='white_name',
            field=models.CharField(blank=True, max_length=500),
        ),
        migrations.AddField(
            model_name='video',
            name='white_slug',
            field=models.SlugField(blank=True, max_length=500),
        ),
        migrations.AddField(
            model_name='video',
            name='year',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),

    ]



# Generated by Django 2.1.3 on 2018-11-05 08:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('agadmator', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='video',
            name='engine',
            field=models.BooleanField(db_index=True, default=False),
        ),
        migrations.AddField(
            model_name='video',
            name='error',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='video',
            name='ignore',
            field=models.BooleanField(db_index=True, default=False),
        ),
        migrations.AddField(
            model_name='video',
            name='stream',
            field=models.BooleanField(db_index=True, default=False),
        ),
    ]



import requests
import re
from lxml.html import fromstring
from .utils import date_from_pgn_date_header, year_from_pgn_date, pgn_similarity_ratio, new_session, parse_pgn
import logging
from django.utils.text import slugify
logger = logging.getLogger(__name__)


def get_game(game_id, session=None, markup=None):
    payload = {}
    if markup is None:
        if session is None:
            session = new_session()
        r = session.get(f'http://www.chessgames.com/perl/chessgame?gid={game_id}', timeout=30)
        r.raise_for_status()
        markup = r.text

    pgn_string = fromstring(markup).get_element_by_id('olga-data').get('pgn').strip()
    game = parse_pgn(pgn_string)
    headers = game.headers

    players = re.findall(r'\/perl\/chessplayer\?pid=(\d+)', markup)
    white_id = int(players[0])
    black_id = int(players[1])

    white_name = headers.get('White', '')
    black_name = headers.get('Black', '')
    payload['white'] = {'chessgames_id': white_id, 'name': white_name, 'slug': slugify(white_name)}
    payload['black'] = {'chessgames_id': black_id, 'name': black_name, 'slug': slugify(black_name)}

    tournament_match = re.search(r'\/perl\/chess\.pl\?tid=(\d+)\">(.*?)<\/a', markup)
    if tournament_match is not None:
        tournament_id, tournament_name = int(tournament_match.group(1)), tournament_match.group(2)
        payload['tournament'] = {'chessgames_id': tournament_id, 'name': tournament_name}

    payload['chessgames_pgn'] = pgn_string
    payload['year'] = year_from_pgn_date(headers.get('Date', ''))
    payload['date'] = date_from_pgn_date_header(headers.get('Date'))
    return payload


def update_video(video, session=None):
    game_id = video.chessgames_id

    if not game_id:
        return

    payload = get_game(game_id, session=session)

    video.date_played = payload.pop('date')
    video.data.update(payload)
    video.save()


def get_best_match(video, session=None):

    if video.extracted_pgn is None:
        return None

    if session is None:
        session = new_session()
    r = session.get(video.chessgames_similar_game_search_url, timeout=30)
    r.raise_for_status()
    markup = r.text

    id_list = re.findall(r'\/perl\/chessgame\?gid=(\d+)', markup)

    logger.info(f'Found {len(id_list)} possible matches for video {video.id}')

    result_id = None

    if len(id_list) > 10:
        video.ignore = True
        video.error = f'Too many possible matches found: {len(id_list)}'
        video.save()
        return None

    for game_id in id_list:

        game_pgn = get_game_pgn(game_id, session=session)

        if pgn_similarity_ratio(game_pgn, video.extracted_pgn) > 0.8:
            result_id = game_id

    video.data['chessgames_success'] = bool(result_id)
    video.data['chessgames_id_list'] = id_list
    if result_id:
        video.chessgames_id = result_id

    video.save()

    return result_id


def get_game_pgn(game_id, session=None):
    if session is None:
        session = requests.Session()

    r = session.get(f'http://www.chessgames.com/perl/chessgame?gid={game_id}')
    r.raise_for_status()
    markup = r.text
    pgn = fromstring(markup).get_element_by_id('olga-data').get('pgn').strip()
    return pgn



from django.http import Http404
from django.shortcuts import render
from django.views import generic
from django.db.models.aggregates import Count
# Create your views here.
from django.db import connection
from django.db.models import Q, F
from django.contrib.postgres.search import SearchQuery, SearchRank
from .models import Video

from collections import namedtuple


def fetchall(cursor):
    """Return all rows from a cursor as a namedtuple"""
    desc = cursor.description
    nt_result = namedtuple('Result', [col[0] for col in desc])
    return [nt_result(*row) for row in cursor.fetchall()]


class IndexView(generic.TemplateView):
    template_name = 'agadmator/index.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['videos'] = (
            Video.objects.all().order_by('-time_published')[:12]
        )

        with connection.cursor() as cursor:
            cursor.execute("""
            SELECT count(*) AS count,
                   name,
                   slug,
                   row_number() OVER (ORDER BY count(*) DESC, name) AS video_ix,
                   row_number() OVER (ORDER BY name) AS name_ix
            FROM (SELECT data->'black'->>'name' AS name, data->'black'->>'slug' AS slug FROM agadmator_video
                  UNION ALL
                  SELECT data->'white'->>'name' AS name, data->'white'->>'slug' AS slug FROM agadmator_video) t
            WHERE name != ''
            GROUP BY name, slug
            ORDER BY count DESC
            """)
            context['players'] = fetchall(cursor)



        with connection.cursor() as cursor:

            cursor.execute("""
            SELECT
                count(*) AS count,
                data->'tournament'->>'name' AS name,
                data->'tournament'->>'chessgames_id' as id,
                row_number() over (ORDER BY data->'tournament'->>'name') as name_ix,
                row_number() over (ORDER BY count(*) DESC ) as videos_ix
            FROM agadmator_video
            WHERE data->'tournament'->>'chessgames_id' NOTNULL
            GROUP BY data->'tournament'->>'name', data->'tournament'->>'chessgames_id'
            ORDER BY count DESC;
            """)

            context['tournaments'] = fetchall(cursor)

        with connection.cursor() as cursor:
            cursor.execute(
                """SELECT date_part('year', date_played)::INT AS year_played,
                    count(*) AS num_videos
                    FROM agadmator_video
        WHERE date_played NOTNULL
        GROUP BY year_played
        ORDER BY year_played""")
            context['years'] = fetchall(cursor)

        with connection.cursor() as cursor:
            cursor.execute(
                """SELECT to_char(time_published, 'YYYY-MM' ) AS month, count(*) AS num_videos
                FROM agadmator_video
                GROUP BY month
                ORDER BY month DESC""")
            context['published_months'] = fetchall(cursor)

        return context


class VideoListView(generic.ListView):
    model = Video
    paginate_by = 21
    template_name = 'agadmator/video_list.html'
    context_object_name = 'videos'
    ordering = ['-time_published']

    def get_queryset(self):
        qs = super().get_queryset()

        if self.request.GET.get('unknown'):
            qs = qs.filter(chessgames_id=None)

        q = self.request.GET.get('q')
        if q:
            query = SearchQuery(q, config='simple')
            qs = qs.filter(search_vector=query).annotate(rank=SearchRank(F('search_vector'), query)).order_by('-rank')

        return qs


class VideoDetailView(generic.DetailView):
    template_name = 'agadmator/video_detail.html'
    model = Video
    context_object_name = 'video'
    slug_field = 'youtube_id'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        return context


class PlayerDetailView(generic.TemplateView):
    template_name = 'agadmator/player_detail.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        id_or_slug = self.kwargs.get('id_or_slug')
        slug = ''
        chessgames_id = None
        try:
            chessgames_id = int(id_or_slug)
        except ValueError:
            slug = id_or_slug

        videos = Video.objects.all().order_by('-time_published')

        if chessgames_id:
            videos = videos.filter(
                Q(data__white__chessgames_id=chessgames_id) | Q(data__black__chessgames_id=chessgames_id)
            )
            context['chessgames_id'] = chessgames_id
        elif slug:
            videos = videos.filter(
                Q(data__white__slug=slug) | Q(data__black__slug=slug)
            )

        context['videos'] = videos

        name = ''
        for v in videos:
            for side in ['black', 'white']:
                if v.data.get(side, {}).get('slug', '') == slug:
                    name = v.data.get(side, {}).get('name', '')
                    break

        context['name'] = name

        return context


class TournamentDetailView(generic.TemplateView):
    template_name = 'agadmator/tournament_detail.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        chessgames_id = self.kwargs.get('chessgames_id')

        videos = Video.objects.raw("""
        SELECT * from agadmator_video
        where data->'tournament'->>'chessgames_id' = '%s'
        ORDER BY time_published DESC 
        """, [chessgames_id])

        if not videos:
            raise Http404()

        context['videos'] = videos
        context['name'] = videos[0].data.get('tournament').get('name')
        context['chessgames_id'] = chessgames_id
        return context


class AboutView(generic.TemplateView):
    template_name = 'agadmator/about.html'


class YearDetailView(generic.TemplateView):
    template_name = 'agadmator/year_detail.html'

    def get_context_data(self, **kwargs):
        year = self.kwargs.get('year')
        context = super().get_context_data(**kwargs)
        context['videos'] = Video.objects.filter(
            game__date__year=year
        )
        return context


class PublishedMonthDetailView(generic.TemplateView):
    template_name = 'agadmator/published_month_detail.html'

    def get_context_data(self, **kwargs):
        year_month = self.kwargs.get('year_month')
        year, month = year_month.split('-')
        context = super().get_context_data(**kwargs)
        context['videos'] = Video.objects.filter(
            time_published__year=year,
            time_published__month=month
        )
        return context



# Generated by Django 2.1.3 on 2018-11-05 16:13

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('agadmator', '0002_auto_20181105_0808'),
    ]

    operations = [
        migrations.RenameField(
            model_name='video',
            old_name='error',
            new_name='error_message',
        ),
    ]



from setuptools import setup, find_packages

setup(
    name="django-agadmator",
    author='Jaco du Plessis',
    author_email='jaco@jacoduplessis.co.za',
    version="0.2.0",
    packages=find_packages(),
    package_data={
        'agadmator': [
            'templates/agadmator/*.html',
            'static/agadmator/*',
            'static/agadmator/chessground/*',
            'static/agadmator/chessground/images/board/*',
            'static/agadmator/chessground/images/pieces/merida/*',
        ]
    },
    install_requires=[
        'django',
        'requests',
        'lxml',
        'cssselect',
        'feedparser',
        'python-chess',
    ]
)



import re
from datetime import datetime
from difflib import SequenceMatcher
from chess.pgn import read_game, StringExporter
from io import StringIO
from requests import Session


def parse_duration(s):
    s = s.strip()

    days, hours, mins, secs, ms = [None] * 5
    m = re.match(
        r'(?:(?:(?:(?P<days>[0-9]+):)?(?P<hours>[0-9]+):)?(?P<mins>[0-9]+):)?(?P<secs>[0-9]+)(?P<ms>\.[0-9]+)?Z?$', s)
    if m:
        days, hours, mins, secs, ms = m.groups()
    else:
        m = re.match(
            r'''(?ix)(?:P?
                (?:
                    [0-9]+\s*y(?:ears?)?\s*
                )?
                (?:
                    [0-9]+\s*m(?:onths?)?\s*
                )?
                (?:
                    [0-9]+\s*w(?:eeks?)?\s*
                )?
                (?:
                    (?P<days>[0-9]+)\s*d(?:ays?)?\s*
                )?
                T)?
                (?:
                    (?P<hours>[0-9]+)\s*h(?:ours?)?\s*
                )?
                (?:
                    (?P<mins>[0-9]+)\s*m(?:in(?:ute)?s?)?\s*
                )?
                (?:
                    (?P<secs>[0-9]+)(?P<ms>\.[0-9]+)?\s*s(?:ec(?:ond)?s?)?\s*
                )?Z?$''', s)
        if m:
            days, hours, mins, secs, ms = m.groups()
        else:
            m = re.match(r'(?i)(?:(?P<hours>[0-9.]+)\s*(?:hours?)|(?P<mins>[0-9.]+)\s*(?:mins?\.?|minutes?)\s*)Z?$', s)
            if m:
                hours, mins = m.groups()
            else:
                return None

    duration = 0
    if secs:
        duration += float(secs)
    if mins:
        duration += float(mins) * 60
    if hours:
        duration += float(hours) * 60 * 60
    if days:
        duration += float(days) * 24 * 60 * 60
    if ms:
        duration += float(ms)
    return duration


def clean_pgn(s):
    s = s.replace('[WhiteElo "?"]\n', '')
    s = s.replace('[BlackElo "?"]\n', '')
    s = s.replace('[Round "?"]\n', '')
    s = s.replace('[EventDate "?"]\n', '')
    return s.strip()


def extract_pgn(text):
    pgn = ''
    m = re.search(r'\n1\.\s?[a-zA-Z]+\d(.*)\n', text, re.MULTILINE)
    if m is not None:
        pgn = m.group(0)

    return pgn.strip()


def date_from_pgn_date_header(date_string):

    try:
        return datetime.strptime(date_string, '%Y.%m.%d').date()
    except ValueError:
        pass

    return None


def parse_pgn(pgn_string):
    return read_game(StringIO(pgn_string))


def normalize_pgn(pgn_string):
    game = parse_pgn(pgn_string)
    exporter = StringExporter(
        columns=None, headers=False, comments=False, variations=False
    )
    return game.accept(exporter).rstrip('* ')


def similarity_ratio(a, b):
    return SequenceMatcher(None, a, b).ratio()


def pgn_similarity_ratio(a, b):
    return similarity_ratio(normalize_pgn(a), normalize_pgn(b))


def year_from_pgn_date(date_string):
    # date_string could look like 1999.??.??
    match = re.search(r'([\d]{4})', date_string)
    return None if match is None else int(match.group(0))


def extract_players(text):
    # first try simple case where A vs B on new line

    m = re.search(r'^([\w \-_()]+)\s?vs\.?\s?([\w \-_()]+)$', text, re.MULTILINE | re.IGNORECASE)
    if m is not None:
        left = m.group(1).strip(' .')
        right = m.group(2).strip(' .')
        return left, right

    m = re.search(r'(\b[A-Z][\w\-]+\s?)+\b\s+(vs|Vs|VS)\.?\s+(\b[A-Z][\w\-]+\s?)+\b', text, re.MULTILINE)
    if m is None:
        return "", ""
    full = m.group(0)
    full = full.split('\n')[0]
    split_string = ''
    splitters = [
        ' vs ',
        ' Vs ',
        ' VS ',
        ' vs. ',
    ]
    for s in splitters:
        if s in full:
            split_string = s
    if not split_string:
        return '', ''
    left, right = full.split(split_string)
    left = left.strip(' .')
    right = right.strip(' .')
    return left, right


def new_session():
    session = Session()
    session.headers.update({
        'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:62.0) Gecko/20100101 Firefox/62.0'
    })
    return session



import feedparser
from celery import shared_task
from celery.utils.log import get_task_logger
from django.utils.timezone import now
from datetime import timedelta

from .models import Video
from agadmator import YOUTUBE_CHANNEL_ID
from .chessgames import get_best_match, update_video
from django.db import connection
from .utils import new_session
import time
logger = get_task_logger(__name__)


@shared_task
def sync_channel_rss_feed():
    rss_url = f'https://www.youtube.com/feeds/videos.xml?channel_id={YOUTUBE_CHANNEL_ID}'

    num_created = 0

    for entry in feedparser.parse(rss_url).entries:
        youtube_id = entry.yt_videoid
        video, created = Video.objects.get_or_create(
            youtube_id=youtube_id,
            defaults={
                'title': entry.title,
                'time_published': entry.published,
                'description': entry.summary
            }
        )
        if created:
            num_created += 1
            video.store_names()
    logger.info(f'Number of videos created: {num_created}')


@shared_task
def sync_channel_and_schedule_match_on_chessgames():
    logger.info('Updating channels over RSS')
    sync_channel_rss_feed()

    new_videos = Video.objects.filter(time_updated__gt=now() - timedelta(minutes=10))

    logger.info(f'New videos: {len(new_videos)}')

    for video in new_videos:
        find_best_chessgames_match_for_video.delay(video.id)

    if len(new_videos) > 0:
        update_search_vectors()
    logger.info('Done')


@shared_task
def find_best_chessgames_match_for_video(video_id):
    video = Video.objects.get(id=video_id)
    game_id = get_best_match(video)
    if game_id:
        video.data['chessgame_id'] = game_id

    video.data['chessgames_match_time'] = now()
    video.save(update_fields=['data'])
    update_video(video)


@shared_task
def find_best_matches_for_all_videos_not_checked_yet():
    videos = Video.objects.filter(data__chessgames_match_time=None)
    for video in videos:
        find_best_chessgames_match_for_video.delay(video.id)
    logger.info('Scheduled {} games for auto matching.'.format(len(videos)))



def update_search_vectors():

    with connection.cursor() as cursor:

        cursor.execute("""
        update agadmator_video v
        set search_vector = document.vector
        from (
        SELECT id,
        setweight(to_tsvector('simple', coalesce(title, '')),'A') ||
        setweight(to_tsvector('simple', coalesce(data->'white'->>'name','') || ' ' || coalesce(data->'black'->>'name','')), 'A') ||
        setweight(to_tsvector('simple', coalesce(description, '')), 'C') ||
        setweight(to_tsvector('simple', coalesce(array_to_string(ARRAY( SELECT jsonb_array_elements_text(data->'tags')),' '), '')),'A') AS vector
        FROM agadmator_video
        ) document
        where v.id = document.id
            
        """)


def find_best_match_and_update_eligible_videos():

    session = new_session()
    videos = Video.objects.filter(
        ignore=False,
        stream=False,
        engine=False,
        chessgames_id=None,
    ).order_by('-time_published')

    for v in videos:
        w, b = v.extracted_players
        if w == "" or b == "":
            v.ignore = True
            v.error = 'Could not extract players'
            v.save()
            continue
        result = False
        try:
            result = get_best_match(v, session=session)
        except Exception as e:
            result = False
            v.error = str(e)
            v.ignore = True
            v.save()
            continue
        finally:
            time.sleep(1)
        if result:
            update_video(v, session=session)
